#!/usr/bin/env node
'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var bcoin = require('bcoin');
var randomBytes = _interopDefault(require('randombytes'));
var ElectrumClient = _interopDefault(require('electrum-client'));
var yargs = require('yargs');
var fs = require('fs');
var path = require('path');
var chalk = _interopDefault(require('chalk'));

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = Object.setPrototypeOf ||
    ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
    function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
}

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function __makeTemplateObject(cooked, raw) {
    if (Object.defineProperty) { Object.defineProperty(cooked, "raw", { value: raw }); } else { cooked.raw = raw; }
    return cooked;
}

/* tslint:disable:max-classes-per-file */
var BadUserPublicKeyError = /** @class */ (function (_super) {
    __extends(BadUserPublicKeyError, _super);
    function BadUserPublicKeyError() {
        var _newTarget = this.constructor;
        var _this = this;
        var trueProto = _newTarget.prototype;
        _this = _super.call(this, 'Invalid user public key') || this;
        // Alternatively use Object.setPrototypeOf if you have an ES6 environment.
        _this.__proto__ = trueProto;
        return _this;
    }
    return BadUserPublicKeyError;
}(Error));
var BadServicePublicKeyError = /** @class */ (function (_super) {
    __extends(BadServicePublicKeyError, _super);
    function BadServicePublicKeyError() {
        var _newTarget = this.constructor;
        var _this = this;
        var trueProto = _newTarget.prototype;
        _this = _super.call(this, 'Invalid service public key') || this;
        // Alternatively use Object.setPrototypeOf if you have an ES6 environment.
        _this.__proto__ = trueProto;
        return _this;
    }
    return BadServicePublicKeyError;
}(Error));
var BadLockTransactionError = /** @class */ (function (_super) {
    __extends(BadLockTransactionError, _super);
    function BadLockTransactionError() {
        var _newTarget = this.constructor;
        var _this = this;
        var trueProto = _newTarget.prototype;
        _this = _super.call(this, 'This locking transaction is not valid for this service') || this;
        // Alternatively use Object.setPrototypeOf if you have an ES6 environment.
        _this.__proto__ = trueProto;
        return _this;
    }
    return BadLockTransactionError;
}(Error));
/* tslint:enable:max-classes-per-file */

/**
 * Returns a Boolean value that indicates if URI has and only has alphanumeric leteral and  '_',  '-', '.' '~'
 * @param str The URI String
 */
function isURISafe(str) {
    var re = /^[a-zA-Z0-9_\-\.\~]*$/;
    return re.test(str);
}
/**
 * Returns a Boolean value indicates if transaction output is in the correct format of Opcode OP_RETURN.
 * @param output Transaction output
 */
function isValidOP_RETURN(output) {
    // Check that output 0 is an OP_RETURN
    if (!output.script.isNulldata()) {
        return false;
    }
    // Check that output 0 contains 0 satoshis
    if (output.value !== 0) {
        return false;
    }
    // Check that output 0 contains exactly 2 opcodes
    if (output.script.length !== 2) {
        return false;
    }
    return true;
}
/**
 * Returns a Boolean value indicates that whether or not the commit transaction is valid
 * @param tx The commit transaction that needs to be verified.
 * @param userPubKey User's public key
 * @param servicePubKey Service provider's public key
 * @param name A name of at most 64 characters composed of URL-safe characters
 * @param locktime An absolute lock time, in blockheight
 */
function verifyCommitTX(tx, userPubKey, servicePubKey, name, locktime) {
    if (tx.outputs.length < 3) {
        return false;
    }
    // Check that output 0 is an OP_RETURN of the correct form
    if (!isValidOP_RETURN(tx.outputs[0])) {
        return false;
    }
    // Check that output 0 contains a 32-byte nonce
    var nonce = tx.outputs[0].script.code[1].data;
    if (nonce.length !== 32) {
        return false;
    }
    // Check that output 1 is sent to the service's address
    var servicePKH = bcoin.crypto.hash160(servicePubKey);
    var serviceAddr = bcoin.address.fromPubkeyhash(servicePKH);
    if (tx.outputs[1].getAddress().toBase58('testnet') !== serviceAddr.toBase58('testnet')) {
        return false;
    }
    // Check that output 2 is a P2SH
    if (!tx.outputs[2].script.isScripthash()) {
        return false;
    }
    // Check that output 2 script is correct
    var redeemScript = genCommitRedeemScript(userPubKey, nonce, name, locktime);
    var scriptHash = bcoin.crypto.hash160(redeemScript.toRaw());
    var p2shAddr = bcoin.address.fromScripthash(scriptHash);
    if (tx.outputs[2].getAddress().toBase58('testnet') !== p2shAddr.toBase58('testnet')) {
        return false;
    }
    return true;
}
/**
 * Returns a Boolean value indicates that whether or not the lock transaction is valid
 * @param tx The lock transaction
 * @param commitTX The corresponding commit transaction.
 * @param servicePubKey service provider's public key
 */
function verifyLockTX(tx, commitTX, servicePubKey) {
    if (tx.outputs.length < 2) {
        return false;
    }
    // Check that input 0 contains a valid pubkey
    var pubKey = getLockTxPubKey(tx);
    if (pubKey === null || !bcoin.crypto.secp256k1.publicKeyVerify(pubKey)) {
        return false;
    }
    // Check that input 0 name data is only 64 bytes in length
    var nameStr = getLockTxName(tx);
    if (nameStr === null) {
        return false;
    }
    if (nameStr.length > 64) {
        return false;
    }
    // Check that output 1 name data contains only URL-safe characters
    if (!isURISafe(nameStr)) {
        return false;
    }
    // Check that output 0 is a P2PKH
    if (!tx.outputs[0].script.isPubkeyhash()) {
        return false;
    }
    // Check that output 0 is sent to the service's address
    var servicePKH = bcoin.crypto.hash160(servicePubKey);
    var serviceAddr = bcoin.address.fromPubkeyhash(servicePKH);
    if (tx.outputs[0].getAddress().toBase58('testnet') !== serviceAddr.toBase58('testnet')) {
        return false;
    }
    // Check that output 1 is a P2SH
    if (!tx.outputs[1].script.isScripthash()) {
        return false;
    }
    // Check that output 1 script is correct
    var locktime = getLockTxTime(tx);
    var redeemScript = genRedeemScript(pubKey, servicePubKey, locktime);
    var scriptHash = bcoin.crypto.hash160(redeemScript.toRaw());
    var p2shAddr = bcoin.address.fromScripthash(scriptHash);
    if (tx.outputs[1].getAddress().toBase58('testnet') !== p2shAddr.toBase58('testnet')) {
        return false;
    }
    // Check that input 0 is a valid commit TX
    if (!verifyCommitTX(commitTX, pubKey, servicePubKey, nameStr, locktime)) {
        return false;
    }
    return true;
}

/**
 * Generate a redeem script, removing a name/key pair from the blockchain.
 * Validates `userPubkey` and `servicePubkey`.
 *
 * @param userPubkey The user's public key.
 * @param servicePubkey The service's public key.
 * @param alocktime An absolute lock time, in blocks.
 */
function genRedeemScript(userPubkey, servicePubkey, alocktime) {
    // Validate user public key
    if (!bcoin.crypto.secp256k1.publicKeyVerify(userPubkey)) {
        throw new BadUserPublicKeyError();
    }
    // Validate service public key
    if (!bcoin.crypto.secp256k1.publicKeyVerify(servicePubkey)) {
        throw new BadServicePublicKeyError();
    }
    var script = new bcoin.script(null);
    script.pushSym('OP_IF');
    //
    // If spending as user, execute this branch
    // Verify that 0 <= current block size - commit block size
    script.pushInt(0);
    script.pushSym('OP_CHECKSEQUENCEVERIFY');
    script.pushSym('OP_DROP');
    // Check the provided user signature
    script.pushData(userPubkey);
    script.pushSym('OP_CHECKSIG');
    script.pushSym('OP_ELSE');
    //
    // Otherwise, if spending as service, execute this branch
    // Verify that alocktime <= current block size
    script.pushInt(alocktime);
    script.pushSym('OP_CHECKLOCKTIMEVERIFY');
    script.pushSym('OP_DROP');
    // Check the provided service signature
    script.pushData(servicePubkey);
    script.pushSym('OP_CHECKSIG');
    script.pushSym('OP_ENDIF');
    script.compile();
    return script;
}
/**
 * Generate a commit redeem script.
 * @param userPubkey The user's public key.
 * @param nonce A 256-bit buffer representing a nonce.
 * @param name A name of at most 64 characters composed of URL-safe characters.
 * @param locktime An absolute lock time, in blocks.
 */
function genCommitRedeemScript(userPubkey, nonce, name, locktime) {
    // Validate user public key
    if (!bcoin.crypto.secp256k1.publicKeyVerify(userPubkey)) {
        throw new BadUserPublicKeyError();
    }
    var script = new bcoin.script();
    // Verify that at least six blocks have passed since commit
    script.pushInt(6);
    script.pushSym('OP_CHECKSEQUENCEVERIFY');
    script.pushSym('OP_DROP');
    //
    // Hash [256-bit nonce + 2-byte locktime (BE) + 1-byte length of name + name]
    // and check against parameters.
    script.pushSym('OP_HASH256');
    var hashData = serializeCommitData(nonce, locktime, name);
    var hash = bcoin.crypto.hash256(hashData);
    script.pushData(hash);
    script.pushSym('OP_EQUALVERIFY');
    // Check user signature
    script.pushData(userPubkey);
    script.pushSym('OP_CHECKSIG');
    script.compile();
    return script;
}
/**
 * Generate a P2SH address from a redeem script.
 * @param redeemScript The script to use.
 */
function genP2shAddr(redeemScript) {
    var outputScript = bcoin.script.fromScripthash(redeemScript.hash160());
    var p2shAddr = bcoin.address.fromScript(outputScript);
    return p2shAddr;
}
/**
 * Encode `nonce`, `locktime`, and `name` into a buffer of length
 * `32 + 4 + 1 + name.length`.
 * @param nonce 256-bit nonce.
 * @param locktime Lock time of at most 500000000 blocks.
 * @param name Name of at most length 64 containing URL-safe characters.
 */
function serializeCommitData(nonce, locktime, name) {
    // Verify nonce size
    if (nonce.length !== 32) {
        throw new Error('Invalid nonce size');
    }
    // Verify lock time is within bounds
    if (locktime > 500000000) {
        throw new Error('Locktime must be less than 500000000 blocks');
    }
    // Verify name is within bounds
    if (name.length > 64) {
        throw new Error('Name is too long');
    }
    // Create a new buffer. Write no
    var outBuf = new Buffer(32 + 4 + 1 + name.length);
    nonce.copy(outBuf);
    outBuf.writeUInt32BE(locktime, 32);
    outBuf.writeUInt8(name.length, 36);
    outBuf.write(name, 37, name.length, 'ascii');
    return outBuf;
}
/**
 * Convert bytestream `data` into human-readable form. The first 32 bytes are
 * the nonce, followed by 32 bytes representing the locktime, then a 4-byte
 * integer representing the name length, then up to 64 bytes representing the
 * name.
 * @param data Binary data to decode.
 */
function deserializeCommitData(data) {
    if (data.length < 38) {
        throw new Error('Invalid commit data');
    }
    var nonce = data.slice(0, 32);
    var locktime = data.readUInt32BE(32);
    var nameLen = data.readUInt8(36);
    var nameRaw = data.slice(37);
    // Verify the specified name length equals the actual length
    if (nameRaw.length !== nameLen) {
        throw new Error('Name has incorrect length');
    }
    var name = nameRaw.toString('ascii');
    return {
        nonce: nonce,
        locktime: locktime,
        name: name,
    };
}
/**
 * Generate a commit transaction.
 * @param coins Array of coins to fund the transaction.
 * @param name Name for name/key pair.
 * @param locktime Absolute locktime
 * @param commitFee Commit fee, in satoshis.
 * @param registerFee Registration fee, in satoshis.
 * @param escrowFee Escrow fee, in satoshis.
 * @param feeRate Fee rate, in satoshis/kilobyte.
 * @param userRing The user's key ring.
 * @param servicePubKey Service public key.
 */
function genCommitTx(coins, name, locktime, commitFee, registerFee, escrowFee, feeRate, userRing, servicePubKey) {
    //
    // Validate name and the service public key
    if (name.length > 64) {
        throw new Error('Name is too long');
    }
    if (!isURISafe(name)) {
        throw new Error('Invalid character(s) in name');
    }
    if (!bcoin.crypto.secp256k1.publicKeyVerify(servicePubKey)) {
        throw new Error('Invalid service public key');
    }
    // Generate a P2SH address from a redeem script, using a random nonce
    var nonce = randomBytes(32);
    var redeemScript = genCommitRedeemScript(userRing.getPublicKey(), nonce, name, locktime);
    var p2shAddr = genP2shAddr(redeemScript);
    // Generate service address from service public key
    var servicePKH = bcoin.crypto.hash160(servicePubKey);
    var serviceAddr = bcoin.address.fromPubkeyhash(servicePKH);
    var lockTx = new bcoin.mtx();
    // Compute total value of coins
    var total = coins.reduce(function (acc, cur) { return acc + cur.value; }, 0);
    for (var _i = 0, coins_1 = coins; _i < coins_1.length; _i++) {
        var coin = coins_1[_i];
        lockTx.addCoin(coin);
    }
    // Compute change amount
    var changeVal = total - (commitFee + registerFee + escrowFee) - (4 * feeRate);
    // Add nonce OP_RETURN as output 0
    var pubkeyDataScript = bcoin.script.fromNulldata(nonce);
    lockTx.addOutput(bcoin.output.fromScript(pubkeyDataScript, 0));
    // Add service upfront fee as output 1
    lockTx.addOutput({
        address: serviceAddr,
        value: commitFee,
    });
    // Add locked fee as output 2
    // Locks up the fee to register, the fee to be put in escrow, and enough for a 4kb tx at current rates
    lockTx.addOutput({
        address: p2shAddr,
        value: registerFee + escrowFee + 4 * feeRate,
    });
    // Add change output as 3
    lockTx.addOutput({
        address: userRing.getAddress(),
        value: changeVal,
    });
    // Add coins as inputs
    for (var i = 0; i < coins.length; ++i) {
        var coin = coins[i];
        lockTx.scriptInput(i, coin, userRing);
    }
    // Each signature is 72 bytes long
    var virtSize = lockTx.getVirtualSize() + coins.length * 72;
    lockTx.subtractIndex(3, Math.ceil(virtSize / 1000 * feeRate));
    // Sign the coins
    for (var i = 0; i < coins.length; ++i) {
        var coin = coins[i];
        lockTx.signInput(i, coin, userRing, bcoin.script.hashType.ALL);
    }
    return lockTx.toTX();
}
/**
 * Generate a lock transaction.
 * @param commitTX The corresponding commit transaction.
 * @param name The name to use.
 * @param upfrontFee The upfront fee in satoshis to use the service, as
 * determined by the service.
 * @param lockedFee Fee incentivizing registrar to provide service, as
 * determined by the service.
 * @param feeRate Fee rate in satoshi/KB.
 * @param userRing The user key ring.
 * @param servicePubKey Service public key.
 * @param locktime Absolute lock time in blocks.
 */
function genLockTx(commitTX, name, upfrontFee, lockedFee, feeRate, userRing, servicePubKey, locktime) {
    //
    // Input validation
    if (locktime > 500000000) {
        throw new Error('Locktime must be less than 500000000 blocks');
    }
    if (name.length > 64) {
        throw new Error('Name is too long');
    }
    if (!isURISafe(name)) {
        throw new Error('Invalid character(s) in name');
    }
    if (!bcoin.crypto.secp256k1.publicKeyVerify(servicePubKey)) {
        throw new Error('Invalid service public key');
    }
    if (!verifyCommitTX(commitTX, userRing.getPublicKey(), servicePubKey, name, locktime)) {
        throw new Error('Invalid commitment tx');
    }
    // Generate a P2SH address from redeem script
    var redeemScript = genRedeemScript(userRing.getPublicKey(), servicePubKey, locktime);
    var p2shAddr = genP2shAddr(redeemScript);
    // Generate address from service public key
    var servicePKH = bcoin.crypto.hash160(servicePubKey);
    var serviceAddr = bcoin.address.fromPubkeyhash(servicePKH);
    var lockTx = bcoin.mtx.fromOptions({
        version: 2,
    });
    lockTx.addTX(commitTX, 2);
    lockTx.setSequence(0, 6);
    var total = commitTX.outputs[2].value;
    var changeVal = total - upfrontFee - lockedFee;
    // Add upfront fee as output 0
    lockTx.addOutput({
        address: serviceAddr,
        value: upfrontFee,
    });
    // Add locked fee as output 1
    lockTx.addOutput({
        address: p2shAddr,
        value: lockedFee,
    });
    // Add change output as 2
    lockTx.addOutput({
        address: userRing.getAddress(),
        value: changeVal,
    });
    var nonce = commitTX.outputs[0].script.code[1].data;
    var hashData = serializeCommitData(nonce, locktime, name);
    var commitRedeemScript = genCommitRedeemScript(userRing.getPublicKey(), nonce, name, locktime);
    var unlockScript = new bcoin.script();
    unlockScript.pushData(hashData);
    unlockScript.pushData(commitRedeemScript.toRaw());
    unlockScript.compile();
    lockTx.inputs[0].script = unlockScript;
    // Add constant for signature
    var virtSize = lockTx.getVirtualSize() + 72;
    // Calculate fee to be paid
    var fee = Math.ceil(virtSize / 1000 * feeRate);
    lockTx.subtractIndex(2, fee);
    // Add signature
    var sig = lockTx.signature(0, commitRedeemScript, total, userRing.getPrivateKey(), bcoin.script.hashType.ALL, 0);
    unlockScript.insertData(0, sig);
    unlockScript.compile();
    return lockTx.toTX();
}
/**
 * Extract metadata from a script (i.e. the nonce, locktime, and name).
 * @param inputScript Script to read from.
 */
function extractCommitMetadata(inputScript) {
    var meta = deserializeCommitData(inputScript.code[1].data);
    return meta;
}
/**
 * Attempt to extract a name from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The name if present, and `null` otherwise.
 */
function getLockTxName(lockTx) {
    try {
        var metadata = extractCommitMetadata(lockTx.inputs[0].script);
        return metadata.name;
    }
    catch (e) {
        return null;
    }
}
/**
 * Attempt to extract a time lock from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The lock time if present, and `null` otherwise.
 */
function getLockTxTime(lockTx) {
    try {
        var metadata = extractCommitMetadata(lockTx.inputs[0].script);
        return metadata.locktime;
    }
    catch (e) {
        return null;
    }
}
/**
 * Attempt to extract a public key from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The public key if present, and `null` otherwise.
 */
function getLockTxPubKey(lockTx) {
    var inputScript = lockTx.inputs[0].script;
    if (inputScript.code.length < 3) {
        return null;
    }
    var encumberScript = bcoin.script.fromRaw(inputScript.code[2].data);
    if (encumberScript.code.length < 7) {
        return null;
    }
    var pubKey = encumberScript.code[6].data;
    return pubKey;
}
/**
 * Generate an unlock transaction.
 * @param lockTx The corresponding lock transaction.
 * @param commitTx The corresponding commit transaction.
 * @param feeRate The fee rate in satoshi/KB.
 * @param service Whether to use the script as service.
 * @param ring The service key ring if `service`, otherwise user key ring.
 * @param otherPubKey The user key ring if `service`, otherwise service key ring.
 */
function genUnlockTx(lockTx, commitTx, feeRate, service, ring, otherPubKey) {
    // Disambiguate ring public key and the other public key
    var servicePubKey = service ? ring.getPublicKey() : otherPubKey;
    var userPubKey = !service ? ring.getPublicKey() : otherPubKey;
    //
    // Input validation
    if (!verifyLockTX(lockTx, commitTx, servicePubKey)) {
        throw new BadLockTransactionError();
    }
    if (!bcoin.crypto.secp256k1.publicKeyVerify(otherPubKey)) {
        throw new Error('Invalid service public key');
    }
    var locktime = getLockTxTime(lockTx);
    if (locktime === null) {
        throw new Error('Could not extract locktime');
    }
    var redeemScript = genRedeemScript(userPubKey, servicePubKey, locktime);
    var val = lockTx.outputs[1].value; // the P2SH output
    var unlockTx = bcoin.mtx.fromOptions({
        version: 2,
    });
    unlockTx.addTX(lockTx, 1);
    var boolVal = service ? 0 : 1;
    if (service) {
        unlockTx.setLocktime(locktime);
    }
    else {
        unlockTx.setSequence(0, 0);
    }
    unlockTx.addOutput({
        address: ring.getAddress(),
        value: val,
    });
    // Generate new script: <sig> <isUser> <redeemScript>
    var unlockScript = new bcoin.script();
    unlockScript.pushData(unlockTx.signature(0, redeemScript, val, ring.getPrivateKey(), bcoin.script.hashType.ALL, 0));
    unlockScript.pushInt(boolVal);
    unlockScript.pushData(redeemScript.toRaw());
    unlockScript.compile();
    unlockTx.inputs[0].script = unlockScript;
    // Compute a fee by multiplying the size by the rate, then account for it
    var virtSize = unlockTx.getVirtualSize();
    var fee = Math.ceil(virtSize / 1000 * feeRate);
    unlockTx.subtractFee(fee);
    // Remake script with the new signature
    var unlockScript2 = new bcoin.script();
    unlockScript2.pushData(unlockTx.signature(0, redeemScript, val, ring.getPrivateKey(), bcoin.script.hashType.ALL, 0));
    unlockScript2.pushInt(boolVal);
    unlockScript2.pushData(redeemScript.toRaw());
    unlockScript2.compile();
    unlockTx.inputs[0].script = unlockScript2;
    return unlockTx.toTX();
}

var TXList = /** @class */ (function () {
    function TXList(txs, spent, heights) {
        if (txs.length !== spent.length || txs.length !== heights.length) {
            throw new Error('Lists of TXs, spent outputs, and heights must be of same length');
        }
        var txMap = {};
        var spentMap = {};
        var heightMap = {};
        var tmpTxids = [];
        for (var i = 0; i < txs.length; ++i) {
            var tx = txs[i];
            var txSpent = spent[i];
            var txHeight = heights[i];
            var hash = bcoin.util.revHex(tx.hash('hex'));
            if (tx.outputs.length !== txSpent.length) {
                throw new Error("Bad outputs for tx " + hash + "; got " + txSpent.length + ", expected " + tx.outputs.length);
            }
            txMap[hash] = tx;
            spentMap[hash] = txSpent;
            heightMap[hash] = txHeight;
            tmpTxids.push(hash);
        }
        this.txs = txMap;
        this.spent = spentMap;
        this.heights = heightMap;
        this.txids = tmpTxids;
    }
    TXList.prototype.getTX = function (txid) {
        if (!this.txs.hasOwnProperty(txid)) {
            throw new Error("Unknown txid '" + txid + "'");
        }
        return this.txs[txid];
    };
    TXList.prototype.getOutputSpent = function (txid, output) {
        if (!this.spent.hasOwnProperty(txid)) {
            throw new Error("Unknown txid '" + txid + "'");
        }
        var txSpent = this.spent[txid];
        if (output < 0 || output >= txSpent.length) {
            throw new Error("Unknown output '" + output + "' for txid '" + txid + "'");
        }
        return txSpent[output];
    };
    TXList.prototype.getHeight = function (txid) {
        if (!this.heights.hasOwnProperty(txid)) {
            throw new Error("Unknown txid '" + txid + "'");
        }
        return this.heights[txid];
    };
    TXList.prototype.getTxids = function () {
        return this.txids;
    };
    return TXList;
}());

var servers = {
	"E-X.not.fyi": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"ELECTRUMX.not.fyi": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"ELEX01.blackpole.online": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"VPS.hsmiths.com": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"bitcoin.freedomnode.com": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"btc.smsys.me": {"pruning":"-","s":"995","version":"1.1"},
	"currentlane.lovebitco.in": {"pruning":"-","t":"50001","version":"1.1"},
	"daedalus.bauerj.eu": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"de01.hamster.science": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"ecdsa.net": {"pruning":"-","s":"110","t":"50001","version":"1.1"},
	"elec.luggs.co": {"pruning":"-","s":"443","version":"1.1"},
	"electrum.akinbo.org": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.antumbra.se": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.be": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.coinucopia.io": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.cutie.ga": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.festivaldelhumor.org": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.hsmiths.com": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.qtornado.com": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum.vom-stausee.de": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrum3.hachre.de": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrumx.bot.nu": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"electrumx.westeurope.cloudapp.azure.com": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"elx01.knas.systems": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"ex-btc.server-on.net": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"helicarrier.bauerj.eu": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"mooo.not.fyi": {"pruning":"-","s":"50012","t":"50011","version":"1.1"},
	"ndnd.selfhost.eu": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"node.arihanc.com": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"node.xbt.eu": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"node1.volatilevictory.com": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"noserver4u.de": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"qmebr.spdns.org": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"raspi.hsmiths.com": {"pruning":"-","s":"51002","t":"51001","version":"1.1"},
	"s2.noip.pl": {"pruning":"-","s":"50102","version":"1.1"},
	"s5.noip.pl": {"pruning":"-","s":"50105","version":"1.1"},
	"songbird.bauerj.eu": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"us.electrum.be": {"pruning":"-","s":"50002","t":"50001","version":"1.1"},
	"us01.hamster.science": {"pruning":"-","s":"50002","t":"50001","version":"1.1"}
};

var testServers = {
	"testnetnode.arihanc.com": {"t":"51001","s":"51002"},
	"testnet1.bauerj.eu": {"t":"51001","s":"51002"},
	"14.3.140.101": {"t":"51001","s":"51002"},
	"testnet.hsmiths.com": {"t":"53011","s":"53012"},
	"electrum.akinbo.org": {"t":"51001","s":"51002"},
	"ELEX05.blackpole.online": {"t":"52011","s":"52002"}
};

var revHex = bcoin.util.revHex;
/**
 * Return an appropriate list of default servers for a given network
 * @param network The current network
 * @returns A list of electrum servers
 */
function getServerList(network) {
    if (network === 'testnet') {
        return JSON.parse(JSON.stringify(testServers));
    }
    else if (network === 'main') {
        return JSON.parse(JSON.stringify(servers));
    }
    else {
        throw new Error("Unknown network '" + network + "'");
    }
}
/**
 * In-place shuffle an array
 *
 * From https://stackoverflow.com/a/6274381/721868
 * @param a The array to shuffle
 * @returns The shuffled array
 */
function shuffle(a) {
    for (var i = a.length - 1; i > 0; i--) {
        var j = Math.floor(Math.random() * (i + 1));
        _a = [a[j], a[i]], a[i] = _a[0], a[j] = _a[1];
    }
    return a;
    var _a;
}
/**
 * Select a server at random from a given list of servers
 * @param serverList The list of servers to select from
 * @returns A tuple of the server's url and ssl port number
 */
function selectRandomServer(serverList) {
    var keys = shuffle(Object.keys(serverList));
    for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
        var randKey = keys_1[_i];
        var randServer = serverList[randKey];
        if (typeof (randServer.s) === 'undefined') {
            continue;
        }
        return [randKey, Number.parseInt(randServer.s)];
    }
    throw new Error('No valid servers');
}
/**
 * Randomly select a server and attempt to connect to it
 * @param network The current network
 * @returns A connected server
 */
function serverConnect(network) {
    return __awaiter(this, void 0, void 0, function () {
        var serverList, i, _a, url, port, ecl, feeRate, e_1;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    serverList = getServerList(network);
                    i = 0;
                    _b.label = 1;
                case 1:
                    if (!(i < 5)) return [3 /*break*/, 9];
                    _a = selectRandomServer(serverList), url = _a[0], port = _a[1];
                    ecl = new ElectrumClient(port, url, 'tls');
                    _b.label = 2;
                case 2:
                    _b.trys.push([2, 6, , 8]);
                    return [4 /*yield*/, ecl.connect()];
                case 3:
                    _b.sent();
                    // Must use protocol >= 1.1
                    return [4 /*yield*/, ecl.server_version('3.0.5', '1.1')];
                case 4:
                    // Must use protocol >= 1.1
                    _b.sent();
                    return [4 /*yield*/, ecl.blockchainEstimatefee(2)];
                case 5:
                    feeRate = _b.sent();
                    if (feeRate < 0) {
                        throw new Error('Server does not support fee estimation');
                    }
                    return [2 /*return*/, ecl];
                case 6:
                    e_1 = _b.sent();
                    return [4 /*yield*/, ecl.close()];
                case 7:
                    _b.sent();
                    delete serverList[url];
                    return [3 /*break*/, 8];
                case 8:
                    ++i;
                    return [3 /*break*/, 1];
                case 9: throw new Error('Could not connect to a server');
            }
        });
    });
}
// TODO: Define a class that manages server connections
/**
 * Get the estimated fee to have a transaction confirmed in 2 blocks in sat/kb
 * @param network The network from which to get info. Currently either 'main' or 'testnet'
 * @returns The estimated fee in sat/kb
 */
function getFeesSatoshiPerKB(network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, feeRate, feeRateSat;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    return [4 /*yield*/, ecl.blockchainEstimatefee(2)];
                case 2:
                    feeRate = _a.sent();
                    if (!(feeRate < 0)) return [3 /*break*/, 4];
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    throw new Error('Bad fee rate');
                case 4:
                    feeRateSat = bcoin.amount.fromBTC(feeRate).toSatoshis(true);
                    return [4 /*yield*/, ecl.close()];
                case 5:
                    _a.sent();
                    return [2 /*return*/, feeRateSat];
            }
        });
    });
}
/**
 * Get the current block height of the specified network
 * @param network The network for which to check the height
 * @returns The current block height
 */
function getBlockHeight(network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, data;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    return [4 /*yield*/, ecl.blockchainHeaders_subscribe()];
                case 2:
                    data = _a.sent();
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    return [2 /*return*/, data.block_height];
            }
        });
    });
}
function addrToScriptHash(addr) {
    var script = bcoin.script.fromAddress(addr);
    var origHash = script.sha256('hex');
    return revHex(origHash);
}
/**
 * Given a target value, generate a list of Coins that provide sufficient funding for this
 * @param addr The controlling address to check
 * @param target The target value to reach
 * @param network The network on which the transaction will occur
 * @returns A list of Coins with total value greater than or equal to target
 */
function fundTx(addr, target, network) {
    return __awaiter(this, void 0, void 0, function () {
        var coins, ecl, scripthash, txs, totalVal, _i, txs_1, tx, rawTx, fullTx, coin;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    coins = [];
                    return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    scripthash = addrToScriptHash(addr);
                    return [4 /*yield*/, ecl.blockchainScripthash_listunspent(scripthash)];
                case 2:
                    txs = _a.sent();
                    if (!(txs.length === 0)) return [3 /*break*/, 4];
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    throw new Error("No unspent txs found for " + addr);
                case 4:
                    // Sort txs by largest value first
                    txs.sort(function (a, b) {
                        return b.value - a.value;
                    });
                    totalVal = 0;
                    _i = 0, txs_1 = txs;
                    _a.label = 5;
                case 5:
                    if (!(_i < txs_1.length)) return [3 /*break*/, 8];
                    tx = txs_1[_i];
                    return [4 /*yield*/, ecl.blockchainTransaction_get(tx.tx_hash)];
                case 6:
                    rawTx = _a.sent();
                    fullTx = bcoin.tx.fromRaw(rawTx, 'hex');
                    coin = bcoin.coin.fromTX(fullTx, tx.tx_pos, tx.height);
                    coins.push(coin);
                    totalVal += tx.value;
                    if (totalVal >= target) {
                        return [3 /*break*/, 8];
                    }
                    _a.label = 7;
                case 7:
                    _i++;
                    return [3 /*break*/, 5];
                case 8: return [4 /*yield*/, ecl.close()];
                case 9:
                    _a.sent();
                    // Error if all utxos checked and still not enough funds
                    if (totalVal < target) {
                        throw new Error('Insufficient funds available');
                    }
                    return [2 /*return*/, coins];
            }
        });
    });
}
/**
 * Get all transactions with a given address as an input or output
 * @param addr The address to query
 * @param network The network about which to get information
 * @returns A TXList containing all of these transactions
 */
function getAllTX(addr, network) {
    return __awaiter(this, void 0, void 0, function () {
        var _this = this;
        var ecl, scripthash, origTxs, confirmedOnly, txs, unspents, outputsSpent, heights;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    scripthash = addrToScriptHash(addr);
                    return [4 /*yield*/, ecl.blockchainScripthash_getHistory(scripthash)];
                case 2:
                    origTxs = _a.sent();
                    confirmedOnly = origTxs.filter(function (data) { return data.height > 0; });
                    return [4 /*yield*/, Promise.all(confirmedOnly.map(function (tx) { return __awaiter(_this, void 0, void 0, function () {
                            var rawTx, fullTx;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0: return [4 /*yield*/, ecl.blockchainTransaction_get(tx.tx_hash)];
                                    case 1:
                                        rawTx = _a.sent();
                                        fullTx = bcoin.tx.fromRaw(rawTx, 'hex');
                                        return [2 /*return*/, fullTx];
                                }
                            });
                        }); }))];
                case 3:
                    txs = _a.sent();
                    unspents = {};
                    return [4 /*yield*/, Promise.all(txs.map(function (tx) { return __awaiter(_this, void 0, void 0, function () {
                            var _this = this;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0: return [4 /*yield*/, Promise.all(tx.outputs.map(function (out, ind) { return __awaiter(_this, void 0, void 0, function () {
                                            var outAddrObj, outAddr, outAddrScripthash, remoteUtxos;
                                            return __generator(this, function (_a) {
                                                switch (_a.label) {
                                                    case 0:
                                                        outAddrObj = out.getAddress();
                                                        if (outAddrObj === null) {
                                                            return [2 /*return*/, false];
                                                        }
                                                        outAddr = outAddrObj.toBase58(network);
                                                        if (!!(outAddr in unspents)) return [3 /*break*/, 2];
                                                        // Due to async nature of await, must add this or there is a race condition
                                                        unspents[outAddr] = {};
                                                        outAddrScripthash = addrToScriptHash(outAddrObj);
                                                        return [4 /*yield*/, ecl.blockchainScripthash_listunspent(outAddrScripthash)];
                                                    case 1:
                                                        remoteUtxos = _a.sent();
                                                        unspents[outAddr] = remoteUtxos.reduce(function (acc, cur) {
                                                            acc[cur.tx_hash + ':' + cur.tx_pos] = true;
                                                            return acc;
                                                        }, {});
                                                        _a.label = 2;
                                                    case 2: return [2 /*return*/, !((tx.txid() + ':' + ind) in unspents[outAddr])];
                                                }
                                            });
                                        }); }))];
                                    case 1: 
                                    // Iterate over each output
                                    return [2 /*return*/, _a.sent()];
                                }
                            });
                        }); }))];
                case 4:
                    outputsSpent = _a.sent();
                    heights = confirmedOnly.map(function (tx) { return tx.height; });
                    return [4 /*yield*/, ecl.close()];
                case 5:
                    _a.sent();
                    return [2 /*return*/, new TXList(txs, outputsSpent, heights)];
            }
        });
    });
}
/**
 * Get a full tx by its txid
 * @param txid The little-endian txid for which to search
 * @param network The network on which the tx took place
 * @returns A TX object for this tx
 * @throws If the txid is not found
 */
function getTX(txid, network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, rawTx, fullTx;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    return [4 /*yield*/, ecl.blockchainTransaction_get(txid)];
                case 2:
                    rawTx = _a.sent();
                    fullTx = bcoin.tx.fromRaw(rawTx, 'hex');
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    return [2 /*return*/, fullTx];
            }
        });
    });
}
/**
 * Publish a signed tx to the network
 * @param tx The raw transaction to publish
 * @param network The network to which to publish
 * @throws If publishing encountered an error
 */
function postTX(tx, network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, rawTx;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    rawTx = tx.toRaw().toString('hex');
                    _a.label = 2;
                case 2:
                    _a.trys.push([2, , 4, 6]);
                    return [4 /*yield*/, ecl.blockchainTransaction_broadcast(rawTx)];
                case 3:
                    _a.sent();
                    return [3 /*break*/, 6];
                case 4: return [4 /*yield*/, ecl.close()];
                case 5:
                    _a.sent();
                    return [7 /*endfinally*/];
                case 6: return [2 /*return*/];
            }
        });
    });
}

/**
 * Function will find a lock transaction in txs
 * @param txid a list of all transactions with a given address as an input
 * or output
 * @param servicePubKey the services public key stored in a buffer
 * @param curHeight the current height of the blockchain (number of blocks)
 * @return IReadonlyNameInfo returns a name, txid, pubkey and the block that the
 * name name expires in
 */
function extractInfo(txs, servicePubKey, curHeight) {
    var map = {};
    for (var _i = 0, _a = txs.getTxids().slice().reverse(); _i < _a.length; _i++) {
        var txid = _a[_i];
        // a transaction in txs list
        var lockTx = txs.getTX(txid);
        // reference to the lockTx 1st inputs previos output reference
        // (does each transaction in the list only have one input?)
        var prevHash = bcoin.util.revHex(lockTx.inputs[0].prevout.hash);
        var ctx = void 0;
        try {
            // check if the lockTx previous transaction is also in TXlist
            ctx = txs.getTX(prevHash);
        }
        catch (e) {
            continue;
        }
        // Is this a valid lock tx or another random kind?
        var valid = verifyLockTX(txs.getTX(txid), ctx, servicePubKey);
        if (!valid) {
            continue;
        }
        // Determine at what block this tx is spendable
        var height = txs.getHeight(txid);
        var period = getLockTxTime(lockTx);
        var expires = period;
        // console.log(expires);
        // Has the P2SH fee been spent yet, signalling a revocation?
        var revoked = txs.getOutputSpent(txid, 1);
        if (revoked) {
            continue;
        }
        var pubKey = getLockTxPubKey(lockTx);
        var data = {
            txid: txid,
            pubKey: pubKey,
            expires: expires,
            invalid: false,
        };
        var name = getLockTxName(lockTx);
        if (map.hasOwnProperty(name) && txs.getHeight(map[name].txid) === height) {
            map[name].invalid = true;
            continue;
        }
        // If this name is already registered, don't replace it
        // Unless the name expired before this tx even existed, in which case include it
        if (!map.hasOwnProperty(name) || map[name].expires < height) {
            map[name] = data;
        }
    }
    // Now return only the non-expired names
    var mapNonExp = {};
    for (var key in map) {
        if (map[key].expires >= curHeight && !map[key].invalid) {
            var _b = map[key], invalid = _b.invalid, dataCpy = __rest(_b, ["invalid"]);
            mapNonExp[key] = dataCpy;
        }
    }
    return mapNonExp;
}

function bech32Encode(pubKey, network) {
    var hrp;
    if (network === 'main') {
        hrp = 'pk';
    }
    else {
        hrp = 'tp';
    }
    return bcoin.utils.bech32.encode(hrp, 0, pubKey);
}
function bech32Decode(data) {
    var decoded = bcoin.utils.bech32.decode(data);
    var pubKey = decoded.hash;
    var hrp = decoded.hrp;
    var network;
    if (hrp === 'pk') {
        network = 'main';
    }
    else if (hrp === 'tp') {
        network = 'testnet';
    }
    else {
        return null;
    }
    return {
        pubKey: pubKey,
        network: network,
    };
}

/* tslint:disable:no-console */
function error(msg) {
    console.error(chalk(templateObject_1 || (templateObject_1 = __makeTemplateObject(["{red Error: ", "}"], ["{red Error: ", "}"])), msg));
    process.exit(1);
    throw new Error('Somehow, exiting the process failed?');
}
function errorUnfoundTx() {
    return error('Could not find the transaction. Check the txid or try again later.');
}
function errorBadHRP() {
    return error('Invalid pubkey HRP');
}
function errorNoFees() {
    return error('Could not fetch fee information. Try again later.');
}
function errorInvalidLock() {
    return error('This is not a valid bitname registration tx');
}
function errorCantPush() {
    return error('There was a problem publishing the tx. Try again later.');
}
function commit(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, wifData, ring, addr, feeRate, err_1, commitFee, registerFee, escrowFee, coins, err_2, commitTx, txidStr, err_3, err_4;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin.keyring.fromSecret(wifData.trim());
                    addr = ring.getAddress();
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, getFeesSatoshiPerKB(net)];
                case 2:
                    feeRate = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_1 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 4:
                    commitFee = 500000;
                    registerFee = 500000;
                    escrowFee = 1000000;
                    _a.label = 5;
                case 5:
                    _a.trys.push([5, 7, , 8]);
                    return [4 /*yield*/, fundTx(addr, commitFee + registerFee + escrowFee + 8 * feeRate, net)];
                case 6:
                    coins = _a.sent();
                    return [3 /*break*/, 8];
                case 7:
                    err_2 = _a.sent();
                    return [2 /*return*/, error('Could not fund the transaction')];
                case 8:
                    _a.trys.push([8, 15, , 16]);
                    commitTx = genCommitTx(coins, argv.name, argv.locktime, commitFee, registerFee, escrowFee, feeRate, ring, servicePubKey);
                    txidStr = chalk(templateObject_2 || (templateObject_2 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin.util.revHex(commitTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 13];
                    _a.label = 9;
                case 9:
                    _a.trys.push([9, 11, , 12]);
                    return [4 /*yield*/, postTX(commitTx, net)];
                case 10:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 12];
                case 11:
                    err_3 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 12: return [3 /*break*/, 14];
                case 13:
                    console.log(txidStr);
                    console.log(commitTx.toRaw().toString('hex'));
                    _a.label = 14;
                case 14: return [3 /*break*/, 16];
                case 15:
                    err_4 = _a.sent();
                    return [2 /*return*/, error('Could not generate transaction: ' + err_4.message)];
                case 16: return [2 /*return*/];
            }
        });
    });
}
function register(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, wifData, ring, feeRate, err_5, upfrontFee, delayFee, commitTX, err_6, lockTx, txidStr, err_7, err_8;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin.keyring.fromSecret(wifData.trim());
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, getFeesSatoshiPerKB(net)];
                case 2:
                    feeRate = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_5 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 4:
                    upfrontFee = 500000;
                    delayFee = 1000000;
                    _a.label = 5;
                case 5:
                    _a.trys.push([5, 7, , 8]);
                    return [4 /*yield*/, getTX(argv.txid, net)];
                case 6:
                    commitTX = _a.sent();
                    return [3 /*break*/, 8];
                case 7:
                    err_6 = _a.sent();
                    return [2 /*return*/, error('Could not find the commit tx')];
                case 8:
                    _a.trys.push([8, 15, , 16]);
                    lockTx = genLockTx(commitTX, argv.name, upfrontFee, delayFee, feeRate, ring, servicePubKey, argv.locktime);
                    txidStr = chalk(templateObject_3 || (templateObject_3 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin.util.revHex(lockTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 13];
                    _a.label = 9;
                case 9:
                    _a.trys.push([9, 11, , 12]);
                    return [4 /*yield*/, postTX(lockTx, net)];
                case 10:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 12];
                case 11:
                    err_7 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 12: return [3 /*break*/, 14];
                case 13:
                    console.log(txidStr);
                    console.log(lockTx.toRaw().toString('hex'));
                    _a.label = 14;
                case 14: return [3 /*break*/, 16];
                case 15:
                    err_8 = _a.sent();
                    return [2 /*return*/, error('Could not generate transaction: ' + err_8.message)];
                case 16: return [2 /*return*/];
            }
        });
    });
}
function revoke(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, wifData, ring, lockTX, err_9, commitTX, err_10, feeRate, err_11, unlockTx, txidStr, err_12;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin.keyring.fromSecret(wifData.trim());
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, getTX(argv.txid, net)];
                case 2:
                    lockTX = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_9 = _a.sent();
                    return [2 /*return*/, errorUnfoundTx()];
                case 4:
                    _a.trys.push([4, 6, , 7]);
                    return [4 /*yield*/, getTX(bcoin.util.revHex(lockTX.inputs[0].prevout.hash), net)];
                case 5:
                    commitTX = _a.sent();
                    return [3 /*break*/, 7];
                case 6:
                    err_10 = _a.sent();
                    return [2 /*return*/, errorUnfoundTx()];
                case 7:
                    if (!verifyLockTX(lockTX, commitTX, servicePubKey)) {
                        return [2 /*return*/, errorInvalidLock()];
                    }
                    _a.label = 8;
                case 8:
                    _a.trys.push([8, 10, , 11]);
                    return [4 /*yield*/, getFeesSatoshiPerKB(net)];
                case 9:
                    feeRate = _a.sent();
                    return [3 /*break*/, 11];
                case 10:
                    err_11 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 11:
                    unlockTx = genUnlockTx(lockTX, commitTX, feeRate, false, ring, servicePubKey);
                    txidStr = chalk(templateObject_4 || (templateObject_4 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin.util.revHex(unlockTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 16];
                    _a.label = 12;
                case 12:
                    _a.trys.push([12, 14, , 15]);
                    return [4 /*yield*/, postTX(unlockTx, net)];
                case 13:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 15];
                case 14:
                    err_12 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 15: return [3 /*break*/, 17];
                case 16:
                    console.log(txidStr);
                    console.log(unlockTx.toRaw().toString('hex'));
                    _a.label = 17;
                case 17: return [2 /*return*/];
            }
        });
    });
}
function serviceSpend(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var net, wifData, ring, servicePubKey, lockTX, err_13, commitTX, err_14, userPubKey, feeRate, err_15, unlockTx, txidStr, err_16;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    net = argv.testnet ? 'testnet' : 'main';
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin.keyring.fromSecret(wifData.trim());
                    servicePubKey = ring.getPublicKey();
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, getTX(argv.txid, net)];
                case 2:
                    lockTX = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_13 = _a.sent();
                    return [2 /*return*/, errorUnfoundTx()];
                case 4:
                    _a.trys.push([4, 6, , 7]);
                    return [4 /*yield*/, getTX(bcoin.util.revHex(lockTX.inputs[0].prevout.hash), net)];
                case 5:
                    commitTX = _a.sent();
                    return [3 /*break*/, 7];
                case 6:
                    err_14 = _a.sent();
                    return [2 /*return*/, errorUnfoundTx()];
                case 7:
                    if (!verifyLockTX(lockTX, commitTX, servicePubKey)) {
                        return [2 /*return*/, errorInvalidLock()];
                    }
                    userPubKey = getLockTxPubKey(lockTX);
                    if (userPubKey === null) {
                        return [2 /*return*/, errorInvalidLock()];
                    }
                    _a.label = 8;
                case 8:
                    _a.trys.push([8, 10, , 11]);
                    return [4 /*yield*/, getFeesSatoshiPerKB(net)];
                case 9:
                    feeRate = _a.sent();
                    return [3 /*break*/, 11];
                case 10:
                    err_15 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 11:
                    unlockTx = genUnlockTx(lockTX, commitTX, feeRate, true, ring, userPubKey);
                    txidStr = chalk(templateObject_5 || (templateObject_5 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin.util.revHex(unlockTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 16];
                    _a.label = 12;
                case 12:
                    _a.trys.push([12, 14, , 15]);
                    return [4 /*yield*/, postTX(unlockTx, net)];
                case 13:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 15];
                case 14:
                    err_16 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 15: return [3 /*break*/, 17];
                case 16:
                    console.log(txidStr);
                    console.log(unlockTx.toRaw().toString('hex'));
                    _a.label = 17;
                case 17: return [2 /*return*/];
            }
        });
    });
}
function allNames(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, addr, txList, curHeight, info, key, encKey;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    addr = bcoin.address.fromPubkeyhash(bcoin.crypto.hash160(servicePubKey));
                    return [4 /*yield*/, getAllTX(addr, net)];
                case 1:
                    txList = _a.sent();
                    return [4 /*yield*/, getBlockHeight(net)];
                case 2:
                    curHeight = _a.sent();
                    info = extractInfo(txList, servicePubKey, curHeight);
                    for (key in info) {
                        if (!info.hasOwnProperty(key)) {
                            continue;
                        }
                        encKey = bech32Encode(info[key].pubKey, net);
                        console.log(chalk(templateObject_6 || (templateObject_6 = __makeTemplateObject(["{green ", "}\n{blue txid} ", "\n{blue pubk} ", "\n"], ["{green ", "}\\n{blue txid} ", "\\n{blue pubk} ", "\\n"])), key, info[key].txid, encKey));
                    }
                    return [2 /*return*/];
            }
        });
    });
}
function keyGen(argv) {
    var net = argv.testnet ? 'testnet' : 'main';
    var ring = bcoin.keyring.generate(net);
    var wif = ring.toSecret(net);
    var pubKeyRaw = ring.getPublicKey();
    var encPK = bech32Encode(pubKeyRaw, net);
    var addr = ring.getAddress().toBase58(net);
    console.log(chalk(templateObject_7 || (templateObject_7 = __makeTemplateObject(["{magenta secret} ", "\n{blue pubk}   ", "\n{blue addr}   ", ""], ["{magenta secret} ", "\\n{blue pubk}   ", "\\n{blue addr}   ", ""])), wif, encPK, addr));
    if (typeof argv.out !== 'undefined') {
        try {
            fs.writeFileSync(path.resolve(argv.out), wif, 'utf8');
        }
        catch (err) {
            return error("Could not write to " + argv.out);
        }
    }
}
function keyInfo(argv) {
    var data;
    try {
        data = fs.readFileSync(path.resolve(argv.file), 'utf8');
    }
    catch (err) {
        return error("Could not read " + argv.file);
    }
    var ring = bcoin.keyring.fromSecret(data.trim());
    var net = ring.network.toString();
    if (net !== 'testnet' && net !== 'main') {
        return error("Unknown network " + net);
    }
    var pubKeyRaw = ring.getPublicKey();
    var encPK = bech32Encode(pubKeyRaw, net);
    var addr = ring.getAddress().toBase58(net);
    console.log(chalk(templateObject_8 || (templateObject_8 = __makeTemplateObject(["{blue pubk} ", "\n{blue addr} ", ""], ["{blue pubk} ", "\\n{blue addr} ", ""])), encPK, addr));
}
function main() {
    /* tslint:disable:no-unused-expression */
    yargs.strict()
        .demandCommand(1, 'Please specify a command')
        .command('commit <servicePubKey> <name> <locktime>', 'commit to intent to register a name', function (yargsObj) {
        return yargsObj
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        })
            .positional('name', {
            type: 'string',
            describe: 'the name to register',
        })
            .positional('locktime', {
            type: 'number',
            describe: 'the block height (up to 499999999) until which the name should be valid',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, commit)
        .command('register <servicePubKey> <txid> <name> <locktime>', 'register a name with the service', function (yargsObj) {
        return yargsObj
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        })
            .positional('txid', {
            type: 'string',
            describe: 'the txid of the commit transaction',
        })
            .positional('name', {
            type: 'string',
            describe: 'the name to register',
        })
            .positional('locktime', {
            type: 'number',
            describe: 'the block height (up to 499999999) until which the name should be valid',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, register)
        .command('revoke <servicePubKey> <txid>', 'revoke a name', function (yargsObj) {
        return yargsObj
            .positional('txid', {
            type: 'string',
            describe: 'the txid of the registering transaction',
        })
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, revoke)
        .command('service-spend <txid>', 'spend locked fee sent to a service you control', function (yargsObj) {
        return yargsObj
            .positional('txid', {
            type: 'string',
            describe: 'the txid of the registering transaction',
        })
            .option('testnet', {
            type: 'boolean',
            describe: 'is this a testnet service?',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, serviceSpend)
        .command('all-names <servicePubKey>', 'get all current names registered with a service', function (yargsObj) {
        return yargsObj
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        });
    }, allNames)
        .command('key-gen', 'generate a new priv/pub keypair', function (yargsObj) {
        return yargsObj
            .option('testnet', {
            type: 'boolean',
            describe: 'whether to generate a testnet key',
        })
            .option('out', {
            alias: 'o',
            type: 'string',
            describe: 'write WIF to a file',
        });
    }, keyGen)
        .command('key-info <file>', 'get information about a private key', function (yargsObj) {
        return yargsObj
            .positional('file', {
            type: 'string',
            describe: 'the WIF file containing the private key',
        });
    }, keyInfo)
        .help()
        .argv;
}
main();
var templateObject_1, templateObject_2, templateObject_3, templateObject_4, templateObject_5, templateObject_6, templateObject_7, templateObject_8;
//# sourceMappingURL=bitname-cli.cjs.js.map
