import { coin as Coin, address as Address, tx as TX } from 'bcoin';
import TXList from './TXList';
/**
 * Get the estimated fee to have a transaction confirmed in 2 blocks in sat/kb
 * @param network The network from which to get info. Currently either 'main' or 'testnet'
 * @returns The estimated fee in sat/kb
 */
declare function getFeesSatoshiPerKB(network: string): Promise<number>;
/**
 * Get the current block height of the specified network
 * @param network The network for which to check the height
 * @returns The current block height
 */
declare function getBlockHeight(network: string): Promise<number>;
/**
 * Given a target value, generate a list of Coins that provide sufficient funding for this
 * @param addr The controlling address to check
 * @param target The target value to reach
 * @param network The network on which the transaction will occur
 * @returns A list of Coins with total value greater than or equal to target
 */
declare function fundTx(addr: Address, target: number, network: string): Promise<Coin[]>;
/**
 * Get all transactions with a given address as an input or output
 * @param addr The address to query
 * @param network The network about which to get information
 * @returns A TXList containing all of these transactions
 */
declare function getAllTX(addr: Address, network: string): Promise<TXList>;
/**
 * Get a full tx by its txid
 * @param txid The little-endian txid for which to search
 * @param network The network on which the tx took place
 * @returns A TX object for this tx
 * @throws If the txid is not found
 */
declare function getTX(txid: string, network: string): Promise<TX>;
/**
 * Publish a signed tx to the network
 * @param tx The raw transaction to publish
 * @param network The network to which to publish
 * @throws If publishing encountered an error
 */
declare function postTX(tx: TX, network: string): Promise<void>;
export { getFeesSatoshiPerKB, getBlockHeight, fundTx, getAllTX, getTX, postTX };
