#!/usr/bin/env node
"use strict";
var __makeTemplateObject = (this && this.__makeTemplateObject) || function (cooked, raw) {
    if (Object.defineProperty) { Object.defineProperty(cooked, "raw", { value: raw }); } else { cooked.raw = raw; }
    return cooked;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var yargs = require("yargs");
var bcoin_1 = require("bcoin");
var txs_1 = require("./lib/txs");
var net_1 = require("./lib/net");
var chain_1 = require("./lib/chain");
var fs = require("fs");
var path = require("path");
var verify_1 = require("./lib/verify");
var utils_1 = require("./lib/utils");
var chalk_1 = require("chalk");
/* tslint:disable:no-console */
function error(msg) {
    console.error(chalk_1.default(templateObject_1 || (templateObject_1 = __makeTemplateObject(["{red Error: ", "}"], ["{red Error: ", "}"])), msg));
    process.exit(1);
    throw new Error('Somehow, exiting the process failed?');
}
function errorUnfoundTx() {
    return error('Could not find the transaction. Check the txid or try again later.');
}
function errorBadHRP() {
    return error('Invalid pubkey HRP');
}
function errorNoFees() {
    return error('Could not fetch fee information. Try again later.');
}
function errorInvalidLock() {
    return error('This is not a valid bitname registration tx');
}
function errorCantPush() {
    return error('There was a problem publishing the tx. Try again later.');
}
function commit(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, wifData, ring, addr, feeRate, err_1, commitFee, registerFee, escrowFee, coins, err_2, commitTx, txidStr, err_3, err_4;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = utils_1.bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin_1.keyring.fromSecret(wifData.trim());
                    addr = ring.getAddress();
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, net_1.getFeesSatoshiPerKB(net)];
                case 2:
                    feeRate = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_1 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 4:
                    commitFee = 500000;
                    registerFee = 500000;
                    escrowFee = 1000000;
                    _a.label = 5;
                case 5:
                    _a.trys.push([5, 7, , 8]);
                    return [4 /*yield*/, net_1.fundTx(addr, commitFee + registerFee + escrowFee + 8 * feeRate, net)];
                case 6:
                    coins = _a.sent();
                    return [3 /*break*/, 8];
                case 7:
                    err_2 = _a.sent();
                    return [2 /*return*/, error('Could not fund the transaction')];
                case 8:
                    _a.trys.push([8, 15, , 16]);
                    commitTx = txs_1.genCommitTx(coins, argv.name, argv.locktime, commitFee, registerFee, escrowFee, feeRate, ring, servicePubKey);
                    txidStr = chalk_1.default(templateObject_2 || (templateObject_2 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin_1.util.revHex(commitTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 13];
                    _a.label = 9;
                case 9:
                    _a.trys.push([9, 11, , 12]);
                    return [4 /*yield*/, net_1.postTX(commitTx, net)];
                case 10:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 12];
                case 11:
                    err_3 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 12: return [3 /*break*/, 14];
                case 13:
                    console.log(txidStr);
                    console.log(commitTx.toRaw().toString('hex'));
                    _a.label = 14;
                case 14: return [3 /*break*/, 16];
                case 15:
                    err_4 = _a.sent();
                    return [2 /*return*/, error('Could not generate transaction: ' + err_4.message)];
                case 16: return [2 /*return*/];
            }
        });
    });
}
function register(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, wifData, ring, feeRate, err_5, upfrontFee, delayFee, commitTX, err_6, lockTx, txidStr, err_7, err_8;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = utils_1.bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin_1.keyring.fromSecret(wifData.trim());
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, net_1.getFeesSatoshiPerKB(net)];
                case 2:
                    feeRate = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_5 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 4:
                    upfrontFee = 500000;
                    delayFee = 1000000;
                    _a.label = 5;
                case 5:
                    _a.trys.push([5, 7, , 8]);
                    return [4 /*yield*/, net_1.getTX(argv.txid, net)];
                case 6:
                    commitTX = _a.sent();
                    return [3 /*break*/, 8];
                case 7:
                    err_6 = _a.sent();
                    return [2 /*return*/, error('Could not find the commit tx')];
                case 8:
                    _a.trys.push([8, 15, , 16]);
                    lockTx = txs_1.genLockTx(commitTX, argv.name, upfrontFee, delayFee, feeRate, ring, servicePubKey, argv.locktime);
                    txidStr = chalk_1.default(templateObject_3 || (templateObject_3 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin_1.util.revHex(lockTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 13];
                    _a.label = 9;
                case 9:
                    _a.trys.push([9, 11, , 12]);
                    return [4 /*yield*/, net_1.postTX(lockTx, net)];
                case 10:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 12];
                case 11:
                    err_7 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 12: return [3 /*break*/, 14];
                case 13:
                    console.log(txidStr);
                    console.log(lockTx.toRaw().toString('hex'));
                    _a.label = 14;
                case 14: return [3 /*break*/, 16];
                case 15:
                    err_8 = _a.sent();
                    return [2 /*return*/, error('Could not generate transaction: ' + err_8.message)];
                case 16: return [2 /*return*/];
            }
        });
    });
}
function revoke(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, wifData, ring, lockTX, err_9, commitTX, err_10, feeRate, err_11, unlockTx, txidStr, err_12;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = utils_1.bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin_1.keyring.fromSecret(wifData.trim());
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, net_1.getTX(argv.txid, net)];
                case 2:
                    lockTX = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_9 = _a.sent();
                    // console.log(err);
                    return [2 /*return*/, errorUnfoundTx()];
                case 4:
                    _a.trys.push([4, 6, , 7]);
                    console.log(lockTX.inputs[0].prevout.hash.toString());
                    return [4 /*yield*/, net_1.getTX(bcoin_1.util.revHex(lockTX.inputs[0].prevout.hash), net)];
                case 5:
                    commitTX = _a.sent();
                    return [3 /*break*/, 7];
                case 6:
                    err_10 = _a.sent();
                    console.log(err_10);
                    return [2 /*return*/, errorUnfoundTx()];
                case 7:
                    if (!verify_1.verifyLockTX(lockTX, commitTX, servicePubKey)) {
                        return [2 /*return*/, errorInvalidLock()];
                    }
                    _a.label = 8;
                case 8:
                    _a.trys.push([8, 10, , 11]);
                    return [4 /*yield*/, net_1.getFeesSatoshiPerKB(net)];
                case 9:
                    feeRate = _a.sent();
                    return [3 /*break*/, 11];
                case 10:
                    err_11 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 11:
                    unlockTx = txs_1.genUnlockTx(lockTX, commitTX, feeRate, false, ring, servicePubKey);
                    txidStr = chalk_1.default(templateObject_4 || (templateObject_4 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin_1.util.revHex(unlockTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 16];
                    _a.label = 12;
                case 12:
                    _a.trys.push([12, 14, , 15]);
                    return [4 /*yield*/, net_1.postTX(unlockTx, net)];
                case 13:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 15];
                case 14:
                    err_12 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 15: return [3 /*break*/, 17];
                case 16:
                    console.log(txidStr);
                    console.log(unlockTx.toRaw().toString('hex'));
                    _a.label = 17;
                case 17: return [2 /*return*/];
            }
        });
    });
}
function serviceSpend(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var net, wifData, ring, servicePubKey, lockTX, err_13, commitTX, err_14, feeRate, err_15, unlockTx, txidStr, err_16;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    net = argv.testnet ? 'testnet' : 'main';
                    wifData = fs.readFileSync(path.resolve(argv.wif), 'utf8');
                    ring = bcoin_1.keyring.fromSecret(wifData.trim());
                    servicePubKey = ring.getPublicKey();
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, net_1.getTX(argv.txid, net)];
                case 2:
                    lockTX = _a.sent();
                    return [3 /*break*/, 4];
                case 3:
                    err_13 = _a.sent();
                    return [2 /*return*/, errorUnfoundTx()];
                case 4:
                    _a.trys.push([4, 6, , 7]);
                    return [4 /*yield*/, net_1.getTX(lockTX.inputs[0].prevout.hash, net)];
                case 5:
                    commitTX = _a.sent();
                    return [3 /*break*/, 7];
                case 6:
                    err_14 = _a.sent();
                    return [2 /*return*/, errorUnfoundTx()];
                case 7:
                    if (!verify_1.verifyLockTX(lockTX, commitTX, servicePubKey)) {
                        return [2 /*return*/, errorInvalidLock()];
                    }
                    _a.label = 8;
                case 8:
                    _a.trys.push([8, 10, , 11]);
                    return [4 /*yield*/, net_1.getFeesSatoshiPerKB(net)];
                case 9:
                    feeRate = _a.sent();
                    return [3 /*break*/, 11];
                case 10:
                    err_15 = _a.sent();
                    return [2 /*return*/, errorNoFees()];
                case 11:
                    unlockTx = txs_1.genUnlockTx(lockTX, commitTX, feeRate, true, ring, servicePubKey);
                    txidStr = chalk_1.default(templateObject_5 || (templateObject_5 = __makeTemplateObject(["{green ", "}"], ["{green ", "}"])), bcoin_1.util.revHex(unlockTx.hash('hex')));
                    if (!argv.push) return [3 /*break*/, 16];
                    _a.label = 12;
                case 12:
                    _a.trys.push([12, 14, , 15]);
                    return [4 /*yield*/, net_1.postTX(unlockTx, net)];
                case 13:
                    _a.sent();
                    console.log(txidStr);
                    return [3 /*break*/, 15];
                case 14:
                    err_16 = _a.sent();
                    return [2 /*return*/, errorCantPush()];
                case 15: return [3 /*break*/, 17];
                case 16:
                    console.log(txidStr);
                    console.log(unlockTx.toRaw().toString('hex'));
                    _a.label = 17;
                case 17: return [2 /*return*/];
            }
        });
    });
}
function allNames(argv) {
    return __awaiter(this, void 0, void 0, function () {
        var decoded, net, servicePubKey, addr, txList, curHeight, info, key, encKey;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    decoded = utils_1.bech32Decode(argv.servicePubKey);
                    if (decoded === null) {
                        return [2 /*return*/, errorBadHRP()];
                    }
                    net = decoded.network;
                    servicePubKey = decoded.pubKey;
                    addr = bcoin_1.address.fromPubkeyhash(bcoin_1.crypto.hash160(servicePubKey));
                    return [4 /*yield*/, net_1.getAllTX(addr, net)];
                case 1:
                    txList = _a.sent();
                    return [4 /*yield*/, net_1.getBlockHeight(net)];
                case 2:
                    curHeight = _a.sent();
                    info = chain_1.extractInfo(txList, servicePubKey, curHeight);
                    for (key in info) {
                        if (!info.hasOwnProperty(key)) {
                            continue;
                        }
                        encKey = utils_1.bech32Encode(info[key].pubKey, net);
                        console.log(chalk_1.default(templateObject_6 || (templateObject_6 = __makeTemplateObject(["{green ", "}\n{blue txid} ", "\n{blue pubk} ", "\n"], ["{green ", "}\\n{blue txid} ", "\\n{blue pubk} ", "\\n"])), key, info[key].txid, encKey));
                    }
                    return [2 /*return*/];
            }
        });
    });
}
function keyGen(argv) {
    var net = argv.testnet ? 'testnet' : 'main';
    var ring = bcoin_1.keyring.generate(net);
    var wif = ring.toSecret(net);
    var pubKeyRaw = ring.getPublicKey();
    var encPK = utils_1.bech32Encode(pubKeyRaw, net);
    var addr = ring.getAddress().toBase58(net);
    console.log(chalk_1.default(templateObject_7 || (templateObject_7 = __makeTemplateObject(["{magenta secret} ", "\n{blue pubk}   ", "\n{blue addr}   ", ""], ["{magenta secret} ", "\\n{blue pubk}   ", "\\n{blue addr}   ", ""])), wif, encPK, addr));
    if (typeof argv.out !== 'undefined') {
        try {
            fs.writeFileSync(path.resolve(argv.out), wif, 'utf8');
        }
        catch (err) {
            return error("Could not write to " + argv.out);
        }
    }
}
function keyInfo(argv) {
    var data;
    try {
        data = fs.readFileSync(path.resolve(argv.file), 'utf8');
    }
    catch (err) {
        return error("Could not read " + argv.file);
    }
    var ring = bcoin_1.keyring.fromSecret(data.trim());
    var net = ring.network.toString();
    if (net !== 'testnet' && net !== 'main') {
        return error("Unknown network " + net);
    }
    var pubKeyRaw = ring.getPublicKey();
    var encPK = utils_1.bech32Encode(pubKeyRaw, net);
    var addr = ring.getAddress().toBase58(net);
    console.log(chalk_1.default(templateObject_8 || (templateObject_8 = __makeTemplateObject(["{blue pubk} ", "\n{blue addr} ", ""], ["{blue pubk} ", "\\n{blue addr} ", ""])), encPK, addr));
}
function main() {
    /* tslint:disable:no-unused-expression */
    yargs
        .strict()
        .demandCommand(1, 'Please specify a command')
        .command('commit <servicePubKey> <name> <locktime>', 'commit to intent to register a name', function (yargsObj) {
        return yargsObj
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        })
            .positional('name', {
            type: 'string',
            describe: 'the name to register',
        })
            .positional('locktime', {
            type: 'number',
            describe: 'how many blocks (up to 65535) to hold the name for',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, commit)
        .command('register <servicePubKey> <txid> <name> <locktime>', 'register a name with the service', function (yargsObj) {
        return yargsObj
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        })
            .positional('txid', {
            type: 'string',
            describe: 'the txid of the commit transaction',
        })
            .positional('name', {
            type: 'string',
            describe: 'the name to register',
        })
            .positional('locktime', {
            type: 'number',
            describe: 'how many blocks (up to 65535) to hold the name for',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, register)
        .command('revoke <servicePubKey> <txid>', 'revoke a name', function (yargsObj) {
        return yargsObj
            .positional('txid', {
            type: 'string',
            describe: 'the txid of the registering transaction',
        })
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, revoke)
        .command('service-spend <txid>', 'spend locked fee sent to a service you control', function (yargsObj) {
        return yargsObj
            .positional('txid', {
            type: 'string',
            describe: 'the txid of the registering transaction',
        })
            .option('testnet', {
            type: 'boolean',
            describe: 'is this a testnet service?',
        })
            .option('wif', {
            alias: 'w',
            type: 'string',
            describe: 'path to WIF file',
            demandOption: true,
        })
            .option('push', {
            type: 'boolean',
            describe: 'whether to push this transaction to the network',
        });
    }, serviceSpend)
        .command('all-names <servicePubKey>', 'get all current names registered with a service', function (yargsObj) {
        return yargsObj
            .positional('servicePubKey', {
            type: 'string',
            describe: 'the public key of the service',
        });
    }, allNames)
        .command('key-gen', 'generate a new priv/pub keypair', function (yargsObj) {
        return yargsObj
            .option('testnet', {
            type: 'boolean',
            describe: 'whether to generate a testnet key',
        })
            .option('out', {
            alias: 'o',
            type: 'string',
            describe: 'write WIF to a file',
        });
    }, keyGen)
        .command('key-info <file>', 'get information about a private key', function (yargsObj) {
        return yargsObj
            .positional('file', {
            type: 'string',
            describe: 'the WIF file containing the private key',
        });
    }, keyInfo)
        .help()
        .argv;
}
main();
var templateObject_1, templateObject_2, templateObject_3, templateObject_4, templateObject_5, templateObject_6, templateObject_7, templateObject_8;
