"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var bcoin_1 = require("bcoin");
var TXList = /** @class */ (function () {
    function TXList(txs, spent, heights) {
        if (txs.length !== spent.length || txs.length !== heights.length) {
            throw new Error('Lists of TXs, spent outputs, and heights must be of same length');
        }
        var txMap = {};
        var spentMap = {};
        var heightMap = {};
        var tmpTxids = [];
        for (var i = 0; i < txs.length; ++i) {
            var tx = txs[i];
            var txSpent = spent[i];
            var txHeight = heights[i];
            var hash = bcoin_1.util.revHex(tx.hash('hex'));
            if (tx.outputs.length !== txSpent.length) {
                throw new Error("Bad outputs for tx " + hash + "; got " + txSpent.length + ", expected " + tx.outputs.length);
            }
            txMap[hash] = tx;
            spentMap[hash] = txSpent;
            heightMap[hash] = txHeight;
            tmpTxids.push(hash);
        }
        this.txs = txMap;
        this.spent = spentMap;
        this.heights = heightMap;
        this.txids = tmpTxids;
    }
    TXList.prototype.getTX = function (txid) {
        if (!this.txs.hasOwnProperty(txid)) {
            throw new Error("Unknown txid '" + txid + "'");
        }
        return this.txs[txid];
    };
    TXList.prototype.getOutputSpent = function (txid, output) {
        if (!this.spent.hasOwnProperty(txid)) {
            throw new Error("Unknown txid '" + txid + "'");
        }
        var txSpent = this.spent[txid];
        if (output < 0 || output >= txSpent.length) {
            throw new Error("Unknown output '" + output + "' for txid '" + txid + "'");
        }
        return txSpent[output];
    };
    TXList.prototype.getHeight = function (txid) {
        if (!this.heights.hasOwnProperty(txid)) {
            throw new Error("Unknown txid '" + txid + "'");
        }
        return this.heights[txid];
    };
    TXList.prototype.getTxids = function () {
        return this.txids;
    };
    return TXList;
}());
exports.default = TXList;
