"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var verify_1 = require("./verify");
var txs_1 = require("./txs");
var bcoin_1 = require("bcoin");
function extractInfo(txs, servicePubKey, curHeight) {
    var map = {};
    for (var _i = 0, _a = txs.getTxids().slice().reverse(); _i < _a.length; _i++) {
        var txid = _a[_i];
        var lockTx = txs.getTX(txid);
        var prevHash = bcoin_1.util.revHex(lockTx.inputs[0].prevout.hash);
        var ctx = void 0;
        try {
            ctx = txs.getTX(prevHash);
        }
        catch (e) {
            continue;
        }
        // Is this a valid lock tx or another random kind?
        var valid = verify_1.verifyLockTX(txs.getTX(txid), ctx, servicePubKey);
        if (!valid) {
            continue;
        }
        // Determine at what block this tx is spendable
        var height = txs.getHeight(txid);
        var period = txs_1.getLockTxTime(lockTx);
        var expires = period;
        // console.log(expires);
        // Has the P2SH fee been spent yet, signalling a revocation?
        var revoked = txs.getOutputSpent(txid, 1);
        if (revoked) {
            continue;
        }
        var pubKey = txs_1.getLockTxPubKey(lockTx);
        var data = {
            txid: txid,
            pubKey: pubKey,
            expires: expires,
            invalid: false,
        };
        var name = txs_1.getLockTxName(lockTx);
        if (map.hasOwnProperty(name) && txs.getHeight(map[name].txid) === height) {
            map[name].invalid = true;
            continue;
        }
        // If this name is already registered, don't replace it
        // Unless the name expired before this tx even existed, in which case include it
        if (!map.hasOwnProperty(name) || map[name].expires < height) {
            map[name] = data;
        }
    }
    // Now return only the non-expired names
    var mapNonExp = {};
    for (var key in map) {
        if (map[key].expires >= curHeight && !map[key].invalid) {
            var _b = map[key], invalid = _b.invalid, dataCpy = __rest(_b, ["invalid"]);
            mapNonExp[key] = dataCpy;
        }
    }
    return mapNonExp;
}
exports.extractInfo = extractInfo;
