"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var bcoin_1 = require("bcoin");
var ElectrumClient = require("electrum-client");
var TXList_1 = require("./TXList");
var revHex = bcoin_1.util.revHex;
// tslint:disable-next-line:no-var-requires
var servers = require('../data/servers.json');
// tslint:disable-next-line:no-var-requires
var testServers = require('../data/servers_testnet.json');
/**
 * Return an appropriate list of default servers for a given network
 * @param network The current network
 * @returns A list of electrum servers
 */
function getServerList(network) {
    if (network === 'testnet') {
        return JSON.parse(JSON.stringify(testServers));
    }
    else if (network === 'main') {
        return JSON.parse(JSON.stringify(servers));
    }
    else {
        throw new Error("Unknown network '" + network + "'");
    }
}
/**
 * In-place shuffle an array
 *
 * From https://stackoverflow.com/a/6274381/721868
 * @param a The array to shuffle
 * @returns The shuffled array
 */
function shuffle(a) {
    for (var i = a.length - 1; i > 0; i--) {
        var j = Math.floor(Math.random() * (i + 1));
        _a = [a[j], a[i]], a[i] = _a[0], a[j] = _a[1];
    }
    return a;
    var _a;
}
/**
 * Select a server at random from a given list of servers
 * @param serverList The list of servers to select from
 * @returns A tuple of the server's url and ssl port number
 */
function selectRandomServer(serverList) {
    var keys = shuffle(Object.keys(serverList));
    for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
        var randKey = keys_1[_i];
        var randServer = serverList[randKey];
        if (typeof (randServer.s) === 'undefined') {
            continue;
        }
        return [randKey, Number.parseInt(randServer.s)];
    }
    throw new Error('No valid servers');
}
/**
 * Randomly select a server and attempt to connect to it
 * @param network The current network
 * @returns A connected server
 */
function serverConnect(network) {
    return __awaiter(this, void 0, void 0, function () {
        var serverList, i, _a, url, port, ecl, feeRate, e_1;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    serverList = getServerList(network);
                    i = 0;
                    _b.label = 1;
                case 1:
                    if (!(i < 5)) return [3 /*break*/, 9];
                    _a = selectRandomServer(serverList), url = _a[0], port = _a[1];
                    ecl = new ElectrumClient(port, url, 'tls');
                    _b.label = 2;
                case 2:
                    _b.trys.push([2, 6, , 8]);
                    return [4 /*yield*/, ecl.connect()];
                case 3:
                    _b.sent();
                    // Must use protocol >= 1.1
                    return [4 /*yield*/, ecl.server_version('3.0.5', '1.1')];
                case 4:
                    // Must use protocol >= 1.1
                    _b.sent();
                    return [4 /*yield*/, ecl.blockchainEstimatefee(2)];
                case 5:
                    feeRate = _b.sent();
                    if (feeRate < 0) {
                        throw new Error('Server does not support fee estimation');
                    }
                    return [2 /*return*/, ecl];
                case 6:
                    e_1 = _b.sent();
                    return [4 /*yield*/, ecl.close()];
                case 7:
                    _b.sent();
                    delete serverList[url];
                    return [3 /*break*/, 8];
                case 8:
                    ++i;
                    return [3 /*break*/, 1];
                case 9: throw new Error('Could not connect to a server');
            }
        });
    });
}
// TODO: Define a class that manages server connections
/**
 * Get the estimated fee to have a transaction confirmed in 2 blocks in sat/kb
 * @param network The network from which to get info. Currently either 'main' or 'testnet'
 * @returns The estimated fee in sat/kb
 */
function getFeesSatoshiPerKB(network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, feeRate, feeRateSat;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    return [4 /*yield*/, ecl.blockchainEstimatefee(2)];
                case 2:
                    feeRate = _a.sent();
                    if (!(feeRate < 0)) return [3 /*break*/, 4];
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    throw new Error('Bad fee rate');
                case 4:
                    feeRateSat = bcoin_1.amount.fromBTC(feeRate).toSatoshis(true);
                    return [4 /*yield*/, ecl.close()];
                case 5:
                    _a.sent();
                    return [2 /*return*/, feeRateSat];
            }
        });
    });
}
exports.getFeesSatoshiPerKB = getFeesSatoshiPerKB;
/**
 * Get the current block height of the specified network
 * @param network The network for which to check the height
 * @returns The current block height
 */
function getBlockHeight(network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, data;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    return [4 /*yield*/, ecl.blockchainHeaders_subscribe()];
                case 2:
                    data = _a.sent();
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    return [2 /*return*/, data.block_height];
            }
        });
    });
}
exports.getBlockHeight = getBlockHeight;
function addrToScriptHash(addr) {
    var script = bcoin_1.script.fromAddress(addr);
    var origHash = script.sha256('hex');
    return revHex(origHash);
}
/**
 * Given a target value, generate a list of Coins that provide sufficient funding for this
 * @param addr The controlling address to check
 * @param target The target value to reach
 * @param network The network on which the transaction will occur
 * @returns A list of Coins with total value greater than or equal to target
 */
function fundTx(addr, target, network) {
    return __awaiter(this, void 0, void 0, function () {
        var coins, ecl, scripthash, txs, totalVal, _i, txs_1, tx, rawTx, fullTx, coin;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    coins = [];
                    return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    scripthash = addrToScriptHash(addr);
                    return [4 /*yield*/, ecl.blockchainScripthash_listunspent(scripthash)];
                case 2:
                    txs = _a.sent();
                    if (!(txs.length === 0)) return [3 /*break*/, 4];
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    throw new Error("No unspent txs found for " + addr);
                case 4:
                    // Sort txs by largest value first
                    txs.sort(function (a, b) {
                        return b.value - a.value;
                    });
                    totalVal = 0;
                    _i = 0, txs_1 = txs;
                    _a.label = 5;
                case 5:
                    if (!(_i < txs_1.length)) return [3 /*break*/, 8];
                    tx = txs_1[_i];
                    return [4 /*yield*/, ecl.blockchainTransaction_get(tx.tx_hash)];
                case 6:
                    rawTx = _a.sent();
                    fullTx = bcoin_1.tx.fromRaw(rawTx, 'hex');
                    coin = bcoin_1.coin.fromTX(fullTx, tx.tx_pos, tx.height);
                    coins.push(coin);
                    totalVal += tx.value;
                    if (totalVal >= target) {
                        return [3 /*break*/, 8];
                    }
                    _a.label = 7;
                case 7:
                    _i++;
                    return [3 /*break*/, 5];
                case 8: return [4 /*yield*/, ecl.close()];
                case 9:
                    _a.sent();
                    // Error if all utxos checked and still not enough funds
                    if (totalVal < target) {
                        throw new Error('Insufficient funds available');
                    }
                    return [2 /*return*/, coins];
            }
        });
    });
}
exports.fundTx = fundTx;
/**
 * Get all transactions with a given address as an input or output
 * @param addr The address to query
 * @param network The network about which to get information
 * @returns A TXList containing all of these transactions
 */
function getAllTX(addr, network) {
    return __awaiter(this, void 0, void 0, function () {
        var _this = this;
        var ecl, scripthash, origTxs, confirmedOnly, txs, unspents, outputsSpent, heights;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    scripthash = addrToScriptHash(addr);
                    return [4 /*yield*/, ecl.blockchainScripthash_getHistory(scripthash)];
                case 2:
                    origTxs = _a.sent();
                    confirmedOnly = origTxs.filter(function (data) { return data.height > 0; });
                    return [4 /*yield*/, Promise.all(confirmedOnly.map(function (tx) { return __awaiter(_this, void 0, void 0, function () {
                            var rawTx, fullTx;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0: return [4 /*yield*/, ecl.blockchainTransaction_get(tx.tx_hash)];
                                    case 1:
                                        rawTx = _a.sent();
                                        fullTx = bcoin_1.tx.fromRaw(rawTx, 'hex');
                                        return [2 /*return*/, fullTx];
                                }
                            });
                        }); }))];
                case 3:
                    txs = _a.sent();
                    unspents = {};
                    return [4 /*yield*/, Promise.all(txs.map(function (tx) { return __awaiter(_this, void 0, void 0, function () {
                            var _this = this;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0: return [4 /*yield*/, Promise.all(tx.outputs.map(function (out, ind) { return __awaiter(_this, void 0, void 0, function () {
                                            var outAddrObj, outAddr, outAddrScripthash, remoteUtxos;
                                            return __generator(this, function (_a) {
                                                switch (_a.label) {
                                                    case 0:
                                                        outAddrObj = out.getAddress();
                                                        if (outAddrObj === null) {
                                                            return [2 /*return*/, false];
                                                        }
                                                        outAddr = outAddrObj.toBase58(network);
                                                        if (!!(outAddr in unspents)) return [3 /*break*/, 2];
                                                        // Due to async nature of await, must add this or there is a race condition
                                                        unspents[outAddr] = {};
                                                        outAddrScripthash = addrToScriptHash(outAddrObj);
                                                        return [4 /*yield*/, ecl.blockchainScripthash_listunspent(outAddrScripthash)];
                                                    case 1:
                                                        remoteUtxos = _a.sent();
                                                        unspents[outAddr] = remoteUtxos.reduce(function (acc, cur) {
                                                            acc[cur.tx_hash + ':' + cur.tx_pos] = true;
                                                            return acc;
                                                        }, {});
                                                        _a.label = 2;
                                                    case 2: return [2 /*return*/, !((tx.txid() + ':' + ind) in unspents[outAddr])];
                                                }
                                            });
                                        }); }))];
                                    case 1: 
                                    // Iterate over each output
                                    return [2 /*return*/, _a.sent()];
                                }
                            });
                        }); }))];
                case 4:
                    outputsSpent = _a.sent();
                    heights = confirmedOnly.map(function (tx) { return tx.height; });
                    return [4 /*yield*/, ecl.close()];
                case 5:
                    _a.sent();
                    return [2 /*return*/, new TXList_1.default(txs, outputsSpent, heights)];
            }
        });
    });
}
exports.getAllTX = getAllTX;
/**
 * Get a full tx by its txid
 * @param txid The little-endian txid for which to search
 * @param network The network on which the tx took place
 * @returns A TX object for this tx
 * @throws If the txid is not found
 */
function getTX(txid, network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, rawTx, fullTx;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    return [4 /*yield*/, ecl.blockchainTransaction_get(txid)];
                case 2:
                    rawTx = _a.sent();
                    fullTx = bcoin_1.tx.fromRaw(rawTx, 'hex');
                    return [4 /*yield*/, ecl.close()];
                case 3:
                    _a.sent();
                    return [2 /*return*/, fullTx];
            }
        });
    });
}
exports.getTX = getTX;
/**
 * Publish a signed tx to the network
 * @param tx The raw transaction to publish
 * @param network The network to which to publish
 * @throws If publishing encountered an error
 */
function postTX(tx, network) {
    return __awaiter(this, void 0, void 0, function () {
        var ecl, rawTx;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, serverConnect(network)];
                case 1:
                    ecl = _a.sent();
                    rawTx = tx.toRaw().toString('hex');
                    _a.label = 2;
                case 2:
                    _a.trys.push([2, , 4, 6]);
                    return [4 /*yield*/, ecl.blockchainTransaction_broadcast(rawTx)];
                case 3:
                    _a.sent();
                    return [3 /*break*/, 6];
                case 4: return [4 /*yield*/, ecl.close()];
                case 5:
                    _a.sent();
                    return [7 /*endfinally*/];
                case 6: return [2 /*return*/];
            }
        });
    });
}
exports.postTX = postTX;
