/// <reference types="node" />
import { tx as TX } from 'bcoin';
/**
 * Returns a Boolean value that indicates if URI has and only has alphanumeric leteral and  '_',  '-', '.' '~'
 * @param str The URI String
 */
declare function isURISafe(str: string): boolean;
/**
 * Returns a Boolean value indicates that whether or not the commit transaction is valid
 * @param tx The commit transaction that needs to be verified.
 * @param userPubKey User's public key
 * @param servicePubKey Service provider's public key
 * @param name A name of at most 64 characters composed of URL-safe characters
 * @param locktime An absolute lock time, in blockheight
 */
declare function verifyCommitTX(tx: TX, userPubKey: Buffer, servicePubKey: Buffer, name: string, locktime: number): boolean;
/**
 * Returns a Boolean value indicates that whether or not the lock transaction is valid
 * @param tx The lock transaction
 * @param commitTX The corresponding commit transaction.
 * @param servicePubKey service provider's public key
 */
declare function verifyLockTX(tx: TX, commitTX: TX, servicePubKey: Buffer): boolean;
export { verifyLockTX, isURISafe, verifyCommitTX };
